import os
import yaml
from gpterm.enums import ThemeMode, VoiceStop, GptModel


class Config:
    DEFAULT_CONFIG_PATH = os.path.expanduser("~/.config/gpterm/config.yaml")
    DEFAULT_IMAGE_STORE_PATH = "/var/tmp/gpterm/generated_images"

    def __init__(self, file_path):
        self.file_path = file_path
        self.color_theme = ThemeMode.dark
        self.use_code_format = False  # toggle this to set code formatting on/off
        self.display_advanced = False
        self.use_voice = True
        self.voice_name = 'Karen'
        self.voice_over = False
        self.voice_stop = VoiceStop.period
        self.image_size = 256
        self.image_view = True
        self.image_store = Config.DEFAULT_IMAGE_STORE_PATH
        self.model = GptModel.davinci
        self.temperature = 0.75

    def load(self):
        if os.path.exists(self.file_path):
            try:
                with open(self.file_path) as fp:
                    loaded_cfg = yaml.safe_load(fp)
                if loaded_cfg:
                    if 'color_theme' in loaded_cfg:
                        self.color_theme = ThemeMode[loaded_cfg['color_theme']]
                    self.use_code_format = loaded_cfg.get('use_code_format', self.use_code_format)
                    self.display_advanced = loaded_cfg.get('display_advanced', self.display_advanced)
                    self.use_voice = loaded_cfg.get('use_voice', self.use_voice)
                    self.voice_name = loaded_cfg.get('voice_name', self.voice_name)
                    self.voice_over = loaded_cfg.get('voice_over', self.voice_over)
                    if 'voice_stop' in loaded_cfg:
                        self.voice_stop = VoiceStop[loaded_cfg['voice_stop']]
                    self.image_size = loaded_cfg.get('image_size', self.image_size)
                    self.image_view = loaded_cfg.get('image_view', self.image_view)
                    self.image_store = loaded_cfg.get('image_store', self.image_store)
                    if 'model' in loaded_cfg:
                        self.model = GptModel[loaded_cfg['model']]
                    self.temperature = loaded_cfg.get('temperature', self.temperature)
            except Exception as e:
                print(f"Error loading {self.file_path}: {e}")

    def save(self):
        yaml_dict = {'color_theme': self.color_theme.name,
                     'use_code_format': self.use_code_format,
                     'display_advanced': self.display_advanced,
                     'use_voice': self.use_voice,
                     'voice_name': self.voice_name,
                     'voice_over': self.voice_over,
                     'voice_stop': self.voice_stop.name,
                     'image_size': self.image_size,
                     'image_view': self.image_view,
                     'image_store': self.image_store,
                     'model': self.model.name,
                     'temperature': self.temperature,
                     }

        cfg_folder = os.path.dirname(self.file_path)
        if not os.path.exists(cfg_folder):
            os.makedirs(cfg_folder)
        with open(self.file_path, 'w') as fp:
            yaml.safe_dump(yaml_dict, fp, default_flow_style=False)



import enum
from typing import NamedTuple


class GptModel(enum.Enum):
    # see: https://beta.openai.com/docs/models/gpt-3
    davinci = "text-davinci-003"
    curie = "text-curie-001"
    babbage = "text-babbage-001"
    ada = "text-ada-001"


class Colors(enum.Enum):
    # using RL_PROMPT_START_IGNORE ('\001') and RL_PROMPT_END_IGNORE ('\002') to get the readline module
    # to ignore the color sequences in prompt when calculating cursor position
    black = '\001\033[90m\002'
    red = '\001\033[91m\002'
    green = '\001\033[1;92m\002'
    yellow = '\001\033[93m\002'
    blue = '\001\033[1;94m\002'
    magenta = '\001\033[1;95m\002'
    cyan = '\001\033[96m\002'
    white = '\001\033[97m\002'
    end = '\001\033[0m\002'


class ThemeMode(enum.Enum):
    dark = 0
    light = 1


class ThemeColors(NamedTuple):
    title: str
    info: str   # shell info color
    cinfo: str  # console info color
    prompt: str
    cmessage: str
    cinput: str
    cresponse: str
    end: str = Colors.end.value


class VoiceStop(enum.Enum):
    period = 0
    newline = 1



#!/usr/bin/env python3

import os
import sys
import math
import time
import datetime
import openai
import argparse
from rich.prompt import Prompt
from rich import print
from rich.console import Console
from rich.syntax import Syntax
from collections import namedtuple
from urllib import request
from gpterm.config import Config
from gpterm.shell import ShellHandler
from gpterm.enums import ThemeColors, Colors, ThemeMode, GptModel, VoiceStop


class GptTerminal:
    def __init__(self, debug=False, theme=None, api_key=None, api_key_path='~/.openai-api-key'):
        self.debug = debug
        self.api_key = api_key
        self.api_key_path = os.path.abspath(os.path.expanduser(os.path.expandvars(api_key_path)))
        self.shell = None
        self.conversation = ""
        self.conversation_formatted = ""
        self.prompt_idx = 0
        self.resp_line = ""
        self.resp_sentence = ""
        self.resp_start = True
        self.in_code_block = False
        self.prompt = ""
        self.prompt_input = ""
        self.first_sentence = True
        self.in_gpt_response = False
        self.abort_response = False
        self.config_file_path = Config.DEFAULT_CONFIG_PATH
        self.cfg = Config(self.config_file_path)
        self.cfg.load()
        if theme:
            self.cfg.color_theme = ThemeMode[theme]
        self.colors = self.get_term_colors()
        self._setup_openai()
        self._setup_gpt()
        self._setup_console()
        self._setup_code_format()
        self._setup_voice()

    def get_commands(self, advanced=False):
        Command = namedtuple('Command', ['advanced', 'setting', 'nargs', 'description'])
        commands = {
            '/help':  Command(False, None, 0, "List available commands"),
            '/exit': Command(False, None, 0, "Exit GPTerm"),
            '/save': Command(False, None, 0, "Save current settings"),
            '/context': Command(False, None, 0, "Print the current chat context (conversation) to screen"),
            '/reset': Command(False, None, 0, "Reset the chat context"),
            '/block': Command(False, None, 0, "Enter a multi-line input"),
            '/image': Command(False, None, 0, "Generate an image from a description using a DALL·E model"),
            '/theme': Command(False, self.cfg.color_theme, 0, "Toggle color theme to match background: light or dark"),
            '/code': Command(False, self.cfg.use_code_format, 0, "Toggle code format on/off (toggling will reset context)"),
            '/voice': Command(False, self.cfg.use_voice, 0, "Toggle voice on/off"),
            '/advanced': Command(False, self.cfg.display_advanced, 0, "Toggle display of advanced commands"),
            '/image-size': Command(True, self.cfg.image_size, 1, "Set the size (pixels x pixels) of generated images. options: 256, 512, 1024"),
            '/image-view': Command(True, self.cfg.image_view, 0, "Toggle to display image after generation in default viewer or not"),
            '/image-store': Command(True, None, 1, "Set the path to store generated images"),
            '/voice-name': Command(True, self.cfg.voice_name, 1, "Set the voice to be used"),
            '/voice-over': Command(True, self.cfg.voice_over, 0, "Toggle voice over highlighting"),
            '/voice-stop': Command(True, self.cfg.voice_stop, 0, "Toggle voice stop: period or newline"),
            '/model': Command(True, self.cfg.model, 1, "GPT-3 Model. Possible options: davinci, curie, babbage, ada"),
            '/temperature': Command(True, self.cfg.temperature, 1, "Provide a value between 0 and 1. Higher for more creative responses. Lower for more standard.")
        }
        if advanced:
            return commands
        else:
            return {key: commands[key] for key in commands if not commands[key].advanced}

    def _setup_gpt(self):
        self.stream = True
        self.max_tokens = self.tokens_per_model()
        self.chars_per_token = 1

    def _setup_code_format(self):
        self.code_format_directive = "\nAny code snippet in your responses must be inside a code block. respond yes if you'll compy"
        self.code_lang = "python"
        self.code_syntax_theme = 'github-dark'  # rich.syntax.DEFAULT_THEME
        self.code_line_numbers = False
        self.code_block_idx = 0

    def _setup_voice(self):
        self.stream_voiced_text = True
        self.stream_text_delay = 0.005

    def toggle_advanced(self):
        self.cfg.display_advanced = not self.cfg.display_advanced
        return self.cfg.display_advanced

    def toggle_voice(self):
        self.cfg.use_voice = not self.cfg.use_voice
        return self.cfg.use_voice

    def toggle_voice_over(self):
        self.cfg.voice_over = not self.cfg.voice_over
        return self.cfg.voice_over

    def toggle_voice_stop(self):
        self.cfg.voice_stop = VoiceStop.period if self.cfg.voice_stop == VoiceStop.newline else VoiceStop.newline
        return self.cfg.voice_stop

    def toggle_image_view(self):
        self.cfg.image_view = not self.cfg.image_view
        return self.cfg.image_view

    def toggle_theme(self):
        self.cfg.color_theme = ThemeMode.dark if self.cfg.color_theme == ThemeMode.light else ThemeMode.light
        self.colors = self.get_term_colors()
        return self.cfg.color_theme

    def toggle_code(self):
        self.cfg.use_code_format = not self.cfg.use_code_format
        return self.cfg.use_code_format

    def _setup_openai(self):
        if self.api_key:
            openai.api_key = self.api_key
        elif os.path.exists(self.api_key_path):
            openai.api_key_path = self.api_key_path

    def _setup_console(self):
        Prompt.prompt_suffix = ''
        self.console = Console()

    def get_term_colors(self):
        dark_theme = ThemeColors(title=Colors.magenta.value, info=Colors.blue.value, cinfo="#6973f6",
                                 prompt=Colors.white.value, cmessage="#C7E9FF", cinput="bold cyan", cresponse="#FFE8F3")
        light_theme = ThemeColors(title=Colors.magenta.value, info=Colors.blue.value, cinfo="bold #0000ff",
                                  prompt=Colors.black.value, cmessage="#191846", cinput="bold #000099", cresponse="bold #ad1f98")
        return dark_theme if self.cfg.color_theme == ThemeMode.dark else light_theme

    def tokens_per_model(self):
        safety_gap = 10
        if self.cfg.model == GptModel.davinci:
            model_max_tokens = 4096
        else:
            model_max_tokens = 2048
        return model_max_tokens - safety_gap

    def calc_max_tokens(self):
        total = self.tokens_per_model()
        self.max_tokens = total - self.text_to_tokens(self.prompt_input)

    def text_to_tokens(self, text, use_tokenizer=False):
        if use_tokenizer:
            # not sure if this runs fast enough on a low-end setup
            try:
                from transformers import GPT2TokenizerFast
                tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")
                return len(tokenizer(text)['input_ids'])
            except ImportError:
                self.print_error("Tokenizer requires module 'transformers'. Install it with: pip install transformers")
        return math.ceil(len(text) / self.chars_per_token)

    def update_max_tokens(self):
        self.calc_max_tokens()
        if self.max_tokens < 0:
            # reset context
            self.console.print(f"[bold red]*** reached max tokens. resetting chat context ***[/]")
            self.reset_context(prompt=self.prompt, submit=False)
        return self.max_tokens

    def reset_context(self, prompt, submit):
        self.prompt = prompt
        current_prompt = f"\n{self.prompt}\n" if self.prompt else ""
        self.add_to_conversation(current_prompt, is_response=False, reset=True)
        if self.cfg.use_code_format:
            self.apply_code_format_directive(prompt=prompt, submit=submit)
        self.prompt_input = self.conversation
        self.calc_max_tokens()
        self.update_shell_prompt()

    def run(self):
        self.shell = ShellHandler()
        self.gpterm_intro = f"{self.colors.title}## GPTerm - Interact with a GPT-3 model via a terminal\n" \
                            f"Type {self.colors.info}/help{self.colors.title} to list available commands.{self.colors.end} " \
                            f"{self.colors.info}/exit{self.colors.title} or {self.colors.info}^C{self.colors.title} to quit.{self.colors.end}\n"
        self.shell.set_gpt_terminal(self)
        self.update_shell_prompt()
        while True:
            self.run_cmdloop()

    def run_cmdloop(self):
        try:
            self.shell.cmdloop()
        except KeyboardInterrupt:
            try:
                if not self.in_gpt_response:
                    response = Prompt.ask("\nExit GPTerm (y/n)? ")
                    if response.lower().startswith('y'):
                        print()
                        sys.exit(0)
                    elif not response:
                        print()
                else:
                    self.in_gpt_response = False
                    print()
                print()
            except KeyboardInterrupt:
                print("\n")

    def update_shell_prompt(self):
        shell_prompt = f"{self.colors.info}🌴 {self.max_tokens} {self.colors.prompt}>{self.colors.end} "
        self.shell.set_shell_prompt(shell_prompt)

    def apply_code_format_directive(self, prompt="", submit=True):
        if not self.cfg.use_code_format:
            return
        self.console.print(f"[{self.colors.cmessage}]{self.code_format_directive}[/]")
        self.prompt = self.code_format_directive + "\n"
        if prompt:
            self.prompt += f"\n{prompt}\n"
        if submit:
            self.submit_prompt(self.prompt)
            self.add_to_conversation("\n", is_response=True, reset=False)
        else:
            self.add_to_conversation(f"{self.prompt}", is_response=False, reset=False)

    def submit_prompt(self, prompt):
        try:
            self.prompt = prompt
            self.add_to_conversation(f"\n{self.prompt}\n", is_response=False)
            self.prompt_input = self.conversation

            if self.debug:
                print(f"[green]{self.prompt_input}[/]", end='')

            self.update_max_tokens()
            self.update_shell_prompt()

            completion = openai.Completion.create(
                headers={"source": "gpterm"},
                engine=self.cfg.model.value,
                prompt=self.prompt_input,
                max_tokens=self.max_tokens,
                n=1,
                temperature=self.cfg.temperature,
                stop=None,
                stream=self.stream,
                # presence_penalty=0,
                # frequency_penalty=0,
            )
            self.handle_completion(completion)
            self.prompt_idx += 1
        except Exception as e:
            self.print_error(e)

    def handle_response_line(self, response, end=False):
        self.resp_line += response
        self.resp_sentence += response
        if self.resp_line.strip() == '```':
            self.code_block_idx = 0
            self.print_chat_response(response.strip() + "\n", force=True)
            self.in_code_block = not self.in_code_block
            self.resp_line = ""
            return

        if not self.in_code_block or response == '``':
            self.print_chat_response(response)
            if self.cfg.use_voice and self.stream_voiced_text and not self.first_sentence:
                time.sleep(self.stream_text_delay)

        has_nl = '\n' in response
        has_fs = '.' in response
        # has_fs = '.' in response or ',' in response
        is_eol = end or has_nl  # end of line
        is_eos = is_eol or has_fs  # end of sentence

        if is_eos:
            self.first_sentence = False
            if is_eol:
                if self.in_code_block:
                    self.print_code_response()
                else:
                    if self.cfg.voice_stop == VoiceStop.newline:
                        self.voice(self.resp_line)
                self.resp_line = ""
            if self.cfg.voice_stop == VoiceStop.period:
                self.voice(self.resp_sentence)
            self.resp_sentence = ""

    def add_to_conversation(self, text, is_response=False, reset=False):
        if reset:
            self.conversation = ""
            self.conversation_formatted = ""
        self.conversation += text
        if is_response:
            if self.resp_start:
                text = self.add_gpt_prefix(text)
            self.conversation_formatted += f"[{self.colors.cresponse}]{text}[/]"
        else:
            if text.strip() != '':
                self.conversation_formatted += f"\n\n[{self.colors.cinput}][Me]: {text.lstrip()}[/]"

    def voice(self, text):
        if not text:
            return
        if not self.cfg.use_voice:
            return
        if self.in_code_block or text == '```' or text == '``':
            return
        line_to_speak = text
        line_to_speak = line_to_speak.replace("\"", "\\\"").replace("`", "")
        if line_to_speak.startswith('-'):
            line_to_speak = f"\\{line_to_speak}"
        try:
            highlight = '-i' if self.cfg.voice_over else ''
            rc = os.system(f'say {highlight} -v {self.cfg.voice_name} \"{line_to_speak}\"')
            if rc != 0:
                self.abort_response = True
        except Exception as e:
            err_msg = f"Failed to produce voiced text: {e}\nYou can disable this feature with /voice"
            self.print_error(err_msg)

    def print_error(self, e):
        self.console.print(f"[bold red]*** got error: {str(e)} ***[/]")

    def reset_response_state(self):
        self.resp_line = ""
        self.resp_sentence = ""
        self.resp_start = True
        self.in_gpt_response = False

    def handle_completion(self, completion):
        if self.stream:
            self.reset_response_state()
            self.first_sentence = True
            self.code_block_idx = 0
            self.abort_response = False
            for idx, obj in enumerate(completion):
                self.in_gpt_response = True
                if self.abort_response:
                    completion.close()
                    self.add_to_conversation("\n")
                    self.reset_response_state()
                    break
                response = obj.choices[0].text
                if self.debug:
                    print(f"[yellow]{response}[/]", end='')

                if idx == 0 and response == '\n':
                    continue

                self.add_to_conversation(response, is_response=True)
                self.handle_response_line(response)
                self.resp_start = False

            if self.resp_line:
                self.handle_response_line('', end=True)

            self.in_gpt_response = False
            print("\n")
        else:
            print(completion.choices[0].text)

    def print_code_response(self):
        if self.resp_line == '\n' and self.code_block_idx == 0:
            return
        self.code_block_idx += 1
        if self.resp_line.endswith('\n'):
            self.resp_line = self.resp_line[:-1]
        syntax = Syntax(self.resp_line, self.code_lang, line_numbers=self.code_line_numbers, theme=self.code_syntax_theme)
        self.console.print(syntax, end='')

    def print_chat_response(self, response, force=False):
        if self.cfg.voice_over and not force:
            return
        if self.resp_start:
            response = self.add_gpt_prefix(response, include_prefix=False)
        self.console.print(f"[{self.colors.cresponse}]{response}[/]", end='')

    def print_info(self, message):
        self.console.print(f"{message}\n")

    def add_gpt_prefix(self, text, include_prefix=True):
        starting_nl = '\n' if text == '``' else ''
        prefix = "[GPT]: " if include_prefix else ""
        text = f"{prefix}{starting_nl}{text}"
        return text

    def submit_image_gen_request(self, prompt):
        size_px = self.cfg.image_size
        self.console.print(f"[{self.colors.cresponse}]Generating Image...[/]")
        try:
            response = openai.Image.create(
                prompt=prompt,
                n=1,
                size=f"{size_px}x{size_px}"
            )
            image_url = response['data'][0]['url']
            self.download_image(image_url=image_url)
        except Exception as e:
            self.print_error(e)

    def download_image(self, image_url):
        if not os.path.exists(self.cfg.image_store):
            os.makedirs(self.cfg.image_store)
        image_filename = "img-" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S%f")[:-3] + ".png"
        image_path = os.path.join(self.cfg.image_store, image_filename)
        request.urlretrieve(image_url, image_path)
        if os.path.exists(image_path):
            self.console.print(f"[{self.colors.cresponse}]Image saved at: {image_path}[/]\n")
            if self.cfg.image_view:
                os.system(f"open {image_path}")
        else:
            self.console.print(f"[{self.colors.cresponse}]Failed to retrieve image[/]\n")


def gpterm_main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--api_key", type=str, help="An OpenAI API key")
    parser.add_argument("--api_key_path", type=str, default='~/.openai-api-key',
                        help="Path to file containing an OpenAI API key")
    parser.add_argument("--theme", type=str, choices=['light', 'dark'],
                        help="Set theme to match background. light or dark")
    args = parser.parse_args()
    debug = False
    gpterm = GptTerminal(debug=debug, theme=args.theme, api_key=args.api_key, api_key_path=args.api_key_path)
    gpterm.run()


if __name__ == "__main__":
    def main():
        gpterm_main()

    main()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="gpterm",
    version="0.2.5",
    author="Jairus",
    author_email="jairus.kope@gmail.com",
    description="A Terminal emulation to interact with OpenAI's GPT models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jairus-kope/gpterm",
    packages=find_packages(),
    install_requires=[
        'openai>=0.26.4',
        'rich>=13.3.1',
        'pyyaml>=6.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        'Operating System :: MacOS :: MacOS X',
    ],
    python_requires='>=3.9.6',
    entry_points={
        "console_scripts": [
            "gpterm = gpterm.gpterm:gpterm_main"
        ]
    },
)



import os
import sys
import readline
import cmd
import enum
from gpterm.enums import ThemeMode, VoiceStop, GptModel
from gpterm.config import Config


class ShellHandler(cmd.Cmd):
    intro = None
    prompt = "> "

    def __init__(self):
        cmd.Cmd.__init__(self)
        self.gpterm = None
        self.loop_index = 0
        self._set_completion_delims()
        self._setup_command_handlers()

    def _set_completion_delims(self):
        delims = readline.get_completer_delims().replace('/', '').replace('-', '')
        readline.set_completer_delims(delims)

    def _setup_command_handlers(self):
        self.cmd_handlers = {
            "/help": self.handle_help,
            "/exit": self.handle_exit,
            "/abort": self.handle_exit,
            "/save": self.handle_save,
            "/reset": self.handle_reset,
            "/context": self.handle_context,
            "/theme": self.handle_theme,
            "/code": self.handle_code,
            "/advanced": self.handle_advanced,
            "/voice": self.handle_voice,
            "/voice-name": self.handle_voice_name,
            "/voice-over": self.handle_voice_over,
            "/voice-stop": self.handle_voice_stop,
            "/model": self.handle_model,
            "/temperature": self.handle_temperature,
            "/block": self.handle_block,
            "/image": self.handle_image,
            "/image-size": self.handle_image_size,
            "/image-view": self.handle_image_view,
            "/image-store": self.handle_image_store,
            "//code_initial": self.handle_code_initial,
        }

    def preloop(self) -> None:
        if self.loop_index == 0:
            self.print_intro()
            if self.gpterm.cfg.use_code_format:
                self.onecmd("//code_initial")
        self.loop_index = 1

    def print_intro(self):
        self.stdout.write(str(self.gpterm.gpterm_intro) + "\n")

    def set_gpt_terminal(self, gpt_terminal):
        self.gpterm = gpt_terminal

    def set_shell_prompt(self, shell_prompt):
        self.prompt = shell_prompt

    def default(self, line):
        if not self.handle_command(line):
            self.gpterm.submit_prompt(line)

    def do_shell(self, line):
        os.system(line)

    def get_multiline(self, instruction="Enter multi-line input"):
        """
        Prompt for multi-line input
        """
        print(f"{instruction}. End with a '.' on a line by itself.")
        lines = []
        while True:
            line = input()
            if line == '.':
                break
            lines.append(line)
        prompt = '\n'.join(lines)
        return prompt

    def emptyline(self):
        return

    def completenames(self, text, *ignored):  # a bit of a hack
        return self.get_command_completions(text)

    def get_command_completions(self, text):
        gpterm_commands = self.gpterm.get_commands(advanced=self.gpterm.cfg.display_advanced)
        gpterm_commands = list(gpterm_commands.keys()) + ['shell']
        if text:
            return [command for command in gpterm_commands if command.startswith(text)]
        else:
            return gpterm_commands

    def do_help(self, line):
        self.default(line)

    def get_voice_name_help(self):
        self.gpterm.console.print(f"Available Voices:")
        self.onecmd("!say -v '?'")
        msg = "Command format: /voice-name <speaker_name>"
        return msg

    def handle_exit(self, _):
        sys.exit(0)

    def handle_help(self, _):
        self.print_intro()
        msg = self.help_message()
        return msg

    def handle_save(self, _):
        self.gpterm.cfg.save()
        msg = "Settings saved to disk"
        return msg

    def handle_reset(self, _):
        self.gpterm.reset_context(prompt="", submit=True)
        msg = f"[bold red]*** Chat context reset ***[/]"
        return msg

    def handle_context(self, _):
        self.gpterm.print_info(f"{self.gpterm.conversation_formatted}")

    def handle_theme(self, _):
        theme = self.gpterm.toggle_theme()
        msg = f"Color theme = {'dark' if theme == ThemeMode.dark else 'light'}"
        return msg

    def handle_code(self, _):
        on = self.gpterm.toggle_code()
        msg = self.handle_reset(_)
        msg += f"\nCode format = {on}"
        return msg

    def handle_advanced(self, _):
        on = self.gpterm.toggle_advanced()
        msg = f"Display advanced commands = {on}"
        return msg

    def handle_voice(self, _):
        on = self.gpterm.toggle_voice()
        msg = f"Voice = {on}"
        return msg

    def handle_voice_name(self, command):
        msg = ""
        if len(command) == 2:
            try:
                rc = os.system(f"say -v {command[1]} ''")
                if rc == 0:
                    self.gpterm.cfg.voice_name = command[1]
                    msg = f"Voice name set to '{command[1]}'"
            except:
                pass
        if not msg:
            msg = self.get_voice_name_help()
        return msg

    def handle_voice_over(self, _):
        on = self.gpterm.toggle_voice_over()
        msg = f"voice over = {on}"
        return msg

    def handle_voice_stop(self, _):
        voice_stop = self.gpterm.toggle_voice_stop()
        msg = f"voice stop = "
        msg += "period" if voice_stop == VoiceStop.period else "newline"
        return msg

    def handle_model(self, command):
        models = [e.name for e in GptModel]
        msg = f"Command format: /model <{'|'.join(models)}>"
        if len(command) == 2 and command[1] in models:
            self.gpterm.cfg.model = GptModel[command[1]]
            self.gpterm.calc_max_tokens()
            self.gpterm.update_shell_prompt()
            msg = f"GPT-3 Model set to '{self.gpterm.cfg.model.name}'"
        return msg

    def handle_temperature(self, command):
        msg = "Command format: /temperature <value between 0-1>"
        if len(command) == 2:
            try:
                val = float(command[1])
                if 0 <= val <= 1:
                    self.gpterm.cfg.temperature = val
                    msg = f"Model temperature set to {self.gpterm.cfg.temperature}"
            except ValueError:
                pass
        return msg

    def handle_block(self, _):
        prompt = self.get_multiline()
        if prompt:
            self.gpterm.submit_prompt(prompt)

    def handle_image(self, _):
        prompt = self.get_multiline(instruction="Enter multi-line input to describe image")
        if prompt:
            self.gpterm.submit_image_gen_request(prompt=prompt)

    def handle_image_size(self, command):
        msg = "Command format: /image-size <256|512|1024>"
        if len(command) == 2:
            try:
                val = int(command[1])
                if val in [256, 512, 1024]:
                    self.gpterm.cfg.image_size = val
                    msg = f"Generated image size set to: {self.gpterm.cfg.image_size} x {self.gpterm.cfg.image_size}"
            except ValueError:
                pass
        return msg

    def handle_image_view(self, _):
        on = self.gpterm.toggle_image_view()
        msg = f"image view = {on}"
        return msg

    def handle_image_store(self, command):
        msg = f"Images store path: {self.gpterm.cfg.image_store}\n"
        msg += f"Command format: /image-store [<path>|reset]"
        if len(command) == 2:
            if command[1] == "reset":
                new_path = Config.DEFAULT_IMAGE_STORE_PATH
            else:
                new_path = os.path.abspath(os.path.expanduser(os.path.expandvars(command[1])))
            msg = f"Failed to set store to '{new_path}'. "
            try:
                if not os.path.exists(new_path):
                    os.makedirs(new_path)
                if os.path.exists(new_path):
                    if os.access(new_path, os.W_OK):
                        self.gpterm.cfg.image_store = new_path
                        msg = f"Images store path: {self.gpterm.cfg.image_store}"
                    else:
                        msg += "Path is not writable"
                else:
                    msg += "Failed to create path"
            except Exception as e:
                msg += f"with error: {e}"
        return msg

    def handle_code_initial(self, _):
        self.gpterm.apply_code_format_directive()

    def handle_command(self, command):
        if not command:
            return False

        if '\n' in command:  # shouldn't happen
            return False

        command = command.strip().split()
        if not command[0].startswith('/'):
            return False

        if command[0] not in self.cmd_handlers:
            return False

        # if len(command) > 2:
        #     self.gpterm.print_info(msg)
        #     return True
        if len(command) > 1:
            all_commands = self.gpterm.get_commands(advanced=True)
            nargs_commands = {key for key in all_commands if all_commands[key].nargs > 0}
            if command[0] not in nargs_commands:
                self.gpterm.print_info(f"Command format: {command[0]}")
                return True

        msg = self.cmd_handlers[command[0]](command)
        if msg:
            self.gpterm.print_info(msg)
        return True

    def help_message(self):
        advanced = self.gpterm.cfg.display_advanced
        msg = f"[{self.gpterm.colors.cinfo}]GPTerm commands:[/]\n"
        gpterm_commands = self.gpterm.get_commands(advanced=advanced)
        key_len = len(max(gpterm_commands.keys(), key=len)) + 1
        for k, v in gpterm_commands.items():
            if v.setting is None:
                state = "\t"
                tab = "\t"
            elif isinstance(v.setting, enum.Enum):
                val_text = v.setting.name
                state = f"= [{self.gpterm.colors.cinfo}]{val_text}[/]"
                tab = '\t' if len(val_text) < 8 else ''
            else:
                val_text = str(v.setting)
                state = f"= {val_text}"
                tab = '\t' if len(val_text) < 8 else ''
            msg += f"{k:<{key_len}} {state}\t{tab} # {v.description}\n"
        if advanced:
            color = "#ff77ff" if self.gpterm.cfg.color_theme == ThemeMode.dark else "#ff00ff"
            msg += "!<shell_command>\t\t # Run a shell command\n"
            msg += f"[{color}]cmd+k[/]\t\t\t\t # Clear the screen\n"
            msg += f"[{color}]ctrl+c[/]\t\t\t\t # Exit GPTerm or abort a GPT response\n"
        return msg.rstrip()
